@@ -32,6 +32,7 @@
 #include	<sys/dir.h>
 #include	<glob.h>
 #include	<fnmatch.h>
+#include	<ctype.h>
 
 /*
  * Read the config file
@@ -211,12 +212,15 @@ void load_partitions(void)
 	}
 	while (fgets(buf, 1024, f)) {
 		int major, minor;
-		char *name;
+		char *name, *mp;
 		buf[1023] = '\0';
 		if (buf[0] != ' ')
 			continue;
-		if (sscanf(buf, " %d %d ", &major, &minor) != 2)
+		major = strtoul(buf, &mp, 10);
+		if (mp == buf || *mp != ' ') 
 			continue;
+		minor = strtoul(mp, NULL, 10);
+
 		name = map_dev(major, minor);
 		if (name) {
 			struct conf_dev *cd;
@@ -227,6 +231,7 @@ void load_partitions(void)
 			cdevlist = cd;
 		}
 	}
+	fclose(f);
 }
 
 
@@ -262,13 +267,14 @@ void arrayline(char *line)
 	mddev_ident_t mi;
 
 	mis.uuid_set = 0;
-	mis.super_minor = -1;
-	mis.level = -10;
-	mis.raid_disks = -1;
-	mis.spare_disks = -1;
+	mis.super_minor = UnSet;
+	mis.level = UnSet;
+	mis.raid_disks = UnSet;
+	mis.spare_disks = UnSet;
 	mis.devices = NULL;
 	mis.devname = NULL;
 	mis.spare_group = NULL;
+	mis.autof = 0;
 
 	for (w=dl_next(line); w!=line; w=dl_next(w)) {
 		if (w[0] == '/') {
@@ -296,7 +302,7 @@ void arrayline(char *line)
 				if (w[12]==0 || endptr[0]!=0 || mis.super_minor < 0) {
 					fprintf(stderr, Name ": invalid super-minor number: %s\n",
 						w);
-					mis.super_minor = -1;
+					mis.super_minor = UnSet;
 				}
 			}
 		} else if (strncasecmp(w, "devices=", 8 ) == 0 ) {
@@ -323,14 +329,42 @@ void arrayline(char *line)
 		} else if (strncasecmp(w, "spares=", 7) == 0 ) {
 			/* for warning if not all spares present */
 			mis.spare_disks = atoi(w+7);
+		} else if (strncasecmp(w, "auto=", 5) == 0 ) {
+			/* whether to create device special files as needed */
+			if (strcasecmp(w+5, "no")==0)
+				mis.autof = 0;
+			else if (strcasecmp(w+5,"yes")==0 || strcasecmp(w+5,"md")==0)
+				mis.autof = -1;
+			else {
+				/* There might be digits, and maybe a hypen, at the end */
+				char *e = w+5 + strlen(w+5);
+				int num = 4;
+				int len;
+				while (e > w+5 && isdigit(e[-1]))
+					e--;
+				if (*e) {
+					num = atoi(e);
+					if (num <= 0) num = 1;
+				}
+				if (e > w+5 && e[-1] == '-')
+					e--;
+				len = e - (w+5);
+				if ((len == 3 && strncasecmp(w+5,"mdp",3)==0) ||
+				    (len == 1 && strncasecmp(w+5,"p",1)==0) ||
+				    (len >= 4 && strncasecmp(w+5,"part",4)==0))
+					mis.autof = num;
+				else 
+					fprintf(stderr, Name ": auto type of \"%s\" ignored for %s\n",
+						w+5, mis.devname?mis.devname:"unlabeled-array");
+			}
 		} else {
 			fprintf(stderr, Name ": unrecognised word on ARRAY line: %s\n",
 				w);
 		}
 	}
 	if (mis.devname == NULL)
-		fprintf(stderr, Name ": ARRAY line with a device\n");
-	else if (mis.uuid_set == 0 && mis.devices == NULL && mis.super_minor < 0)
+		fprintf(stderr, Name ": ARRAY line with no device\n");
+	else if (mis.uuid_set == 0 && mis.devices == NULL && mis.super_minor == UnSet)
 		fprintf(stderr, Name ": ARRAY line %s has no identity information.\n", mis.devname);
 	else {
 		mi = malloc(sizeof(*mi));
@@ -417,6 +451,7 @@ void load_conffile(char *conffile)
 		free_line(line);
 	}
     
+	fclose(f);
 
 /*    printf("got file\n"); */
 }
@@ -450,7 +485,7 @@ mddev_dev_t conf_get_devs(char *conffile)
 	struct conf_dev *cd;
 	int flags = 0;
 	static mddev_dev_t dlist = NULL;
-	int i;
+	unsigned int i;
 
 	while (dlist) {
 		mddev_dev_t t = dlist;
